import logging
import sys

from flask.logging import default_handler

__all__ = []


class Logger:
    @staticmethod
    def configure() -> None:
        # Set default logging to a simple timestamp and message
        # default_handler is StreamHandler(stream=sys.stderr)
        formatter = logging.Formatter("[%(asctime)s] [%(levelname)s] %(message)s")
        default_handler.setStream(sys.stderr)
        default_handler.setLevel(logging.INFO)
        default_handler.setFormatter(formatter)

        # On the "stderr" logger we log info, warning and errors but not requests
        logger = logging.getLogger("stderr")
        handler = logging.StreamHandler(stream=sys.stderr)
        handler.setFormatter(formatter)
        Logger._clear_handlers(logger)
        logger.addHandler(handler)
        logger.setLevel(logging.INFO)

        # On the "request" logger, we log only the requests
        formatter = logging.Formatter("[%(asctime)s] - %(message)s")
        handler = logging.StreamHandler(stream=sys.stdout)
        handler.setFormatter(formatter)
        logger = logging.getLogger("request")
        Logger._clear_handlers(logger)
        logger.addHandler(handler)
        logger.setLevel(logging.DEBUG)

    @staticmethod
    def disable() -> None:
        # Remove the FileHandlers from the Gunicorn logger
        # to prevent double logging
        for name in ["gunicorn.error", "gunicorn.access"]:
            logger = logging.getLogger(name)
            for handler in logger.handlers:
                if handler._gunicorn:
                    logger.removeHandler(handler)

        # Delete the werkzeug logger
        logger = logging.getLogger("werkzeug")
        Logger._clear_handlers(logger)

    @staticmethod
    def log_request(request: object, response: object) -> None:
        # TODO: limit size of path writtent to logs
        url = request.path
        if request.args:
            url = request.full_path
        header = request.headers.get("User-Agent", "")
        message = f'{request.remote_addr} - "{request.method} {url}" {response.status_code} - {header}'

        logger = logging.getLogger("request")
        logger.info(message)

    @staticmethod
    def _clear_handlers(logger) -> None:
        # Clear handlers before adding new handlers
        # This prevents multiple handlers and multiple
        # printing duing unit tests
        for handler in logger.handlers:
            logger.removeHandler(handler)



import json

import html5lib

from flask.testing import FlaskClient
from werkzeug.test import TestResponse

__all__ = []


def _build_query_string(query):
    if not query:
        return ""
    if isinstance(query, str):
        return query
    return "&".join(f"{k}={v}" for k, v in sorted(query.items()))


class MockSession:
    def __init__(self, client):
        self._client = client

    def set(self, key, value):
        with self._client.session_transaction() as client_session:
            client_session[key] = value

    # FIXME: also do a __setitem__ ?
    def __getitem__(self, key):
        return self._items[key]

    @property
    def _items(self):
        with self._client.session_transaction() as client_session:
            return dict(client_session.items())

    def items(self):
        return self._items

    def clear(self):
        with self._client.session_transaction() as client_session:
            client_session.clear()


class ResponseEx(TestResponse):
    @classmethod
    def cast(cls, source: TestResponse):
        assert isinstance(source, TestResponse)
        source.__class__ = cls
        assert isinstance(source, TestResponse)
        return source

    def expect(self, status_code=200, content_type=None):
        """
        Assert the results from a call_method call
        """
        # Verify status code
        if status_code:
            message = f"Expected client to return HTTP code {status_code}, received {self.status_code} instead"
            assert status_code == self.status_code, message

        # Set a content-type if this was a post
        if not content_type and self.request.method == "POST":
            if self.request.content_type == "application/json":
                content_type = "application/json"

        # Get the real data
        data = self.data
        try:
            data = self.json
            if data is not None and not content_type:
                content_type = "application/json"
        except json.decoder.JSONDecodeError:
            if not content_type:
                content_type = "text/html"

        if content_type == "*":
            pass
        elif content_type:
            # switch content-type on error
            if self.status_code >= 400 and content_type == "application/json":
                content_type = "application/problem+json"

            # Verify content_type
            content = self.content_type.split(";")[0]
            message = f"Expecting '{content_type}' as content-type but got '{content}'"
            assert content_type == content, message
        else:
            assert data is None, "A content-type is required when response data is not empty"

        # Verify contents
        if content_type == "text/html":
            html5parser = html5lib.HTMLParser(strict=True)
            html5parser.parse(str(self.data.decode("UTF-8")))

        return self


class MockClient(FlaskClient):
    def __init__(self: object, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._default_headers = {}

    # pylint: disable=arguments-differ
    def get(self, url, **kwargs):
        return self.request(url, method="GET", **kwargs)

    # pylint: disable=arguments-differ
    def post(self, url, **kwargs):
        return self.request(url, method="POST", **kwargs)

    def request(self, url, method="GET", session=None, query=None, headers=None, **kwargs):
        """
        Simple request with parameters, without test assertions

        :param url: URL to call
        :param method: Call method, "GET by default
        :param params: Additional GET params to add
        :returns: result from client call method
        """

        # Build a query string
        query_string = _build_query_string(query)
        url = f"{url}?{query_string}"

        # Create or edit the session
        for k, v in (session or {}).items():
            self.session.set(k, v)

        headers = self._default_headers | (headers or {})

        if method == "GET":
            response = super().get(url, headers=headers, **kwargs)
        elif method == "POST":
            response = super().post(url, headers=headers, **kwargs)
        else:
            assert False, f"call method '{method}' not implemented"
        return ResponseEx.cast(response)

    @property
    def session(self):
        return MockSession(self)

    @property
    def default_headers(self):
        return self._default_headers

    @default_headers.setter
    def default_headers(self, value):
        self._default_headers = value



from werkzeug import exceptions as exception

__all__ = [
    "HTTPException",
    "NoContent",
    "BadRequest",
    "Unauthorized",
    "Forbidden",
    "NotFound",
    "FailedDependency",
    "InternalServerError",
    "NotImplemented",
    "ServiceUnavailable",
]

# Notes on exceptions handling and errors:
# https://werkzeug.palletsprojects.com/en/master/exceptions/
# https://tools.ietf.org/html/rfc7807
HTTPException = exception.HTTPException

# 204 No Content
class NoContent(HTTPException):
    code = 204
    name = "No Content"


# 400 Bad Request
BadRequest = exception.BadRequest

# 401 Unauthorized
Unauthorized = exception.Unauthorized

# 403 Forbidden
Forbidden = exception.Forbidden

# 404 NotFound
NotFound = exception.NotFound

# 424 FailedDependency
FailedDependency = exception.FailedDependency

# 500 InternalServerError
InternalServerError = exception.InternalServerError

# 501 NotImplemented
# pylint: disable=redefined-builtin
NotImplemented = exception.NotImplemented

# 503 ServiceUnavailable
ServiceUnavailable = exception.ServiceUnavailable


class HTTPExceptions:
    pass


# Copy all symbols above into HTTPExceptions
for x in __all__:
    setattr(HTTPExceptions, x, locals()[x])



import functools
import typing as t

from attr import define, field
from flask import current_app, has_app_context

from ..mixins import LoggerProperties

__all__ = ["client"]

g_clients = {}


def client(name: str = None) -> t.Callable:
    fn = None
    if callable(name) and hasattr(name, "__name__"):
        fn = name
        name = fn.__name__

    def register_client(fn, name=None):
        if has_app_context():
            assert False, "flasket.client.__init__: Application has a context"
        else:
            g_clients[name] = fn

    @functools.wraps(fn)
    def wrapped(fn):
        nonlocal name

        register_client(fn, name)
        return fn

    if fn:
        register_client(fn, name)
    return wrapped


@define(kw_only=True)
class ClientFactory(LoggerProperties):
    _flasket = field()

    def __getattr__(self, name):
        self.logger.info(f"Creating client '{name}'...")
        fn = g_clients.get(name)
        if fn is None:
            self.logger.error(f"Client '{name}' is not registered.")
            raise TypeError(f"Client '{name}' does not exist")

        try:
            retval = fn(app=self._flasket, name=name)
            setattr(self.__class__, name, property(fget=lambda self: retval))
            return retval
        except Exception:
            self.logger.error(f"Client '{name}' could not be loaded.")
            raise



import functools
import os

from flask import Flask, Blueprint, send_from_directory

from attrs import define, field

from . import Handler, handler


@handler
@define(kw_only=True)
class StaticHandler(Handler):
    _static_rootpath: str = field(init=False, default=None)

    @property
    def prefix(self) -> str:
        # FIXME We don't add the trailing slash
        return ""

    def __attrs_post_init__(self) -> None:
        # Search for ./static, then for ./htdocs
        rootpath = self.rootpath
        if os.path.exists(filepath := os.path.join(rootpath, "static")):
            self._static_rootpath = filepath
            self.logger.info(f"Reading static files from '{filepath}' directory.")
        elif os.path.exists(filepath := os.path.join(rootpath, "htdocs")):
            self._static_rootpath = filepath
            self.logger.info(f"Reading static files from '{filepath}' directory.")
        else:
            self._static_rootpath = os.path.join(rootpath, "static")
            self.logger.warning("Neither 'static' nor 'htdocs' directories in root path exist.")

        # Use None for static_folder to prevent adding a automatic "/<filename>" route
        ## TODO: Allow serving of html files via the template rendering module
        self._flask = Flask(__name__, root_path=self._static_rootpath, static_folder=None, template_folder=None)
        blueprint = Blueprint(
            "static",
            __name__,
            root_path=self._static_rootpath,
            static_folder=None,
            template_folder=None,
        )
        blueprint.add_url_rule("/<path:path>", view_func=self.send_file)
        self.flask.register_blueprint(blueprint)

        # FIXME: Allow a cfg option 'redirect' from '/' to '/app/'
        redirect_fn = functools.partial(self.redirect, location="/app/", code=308)
        redirect_fn.__name__ = "redirect"
        blueprint = Blueprint(
            "redirect",
            __name__,
            root_path=self.rootpath,
            static_folder=None,
            template_folder=None,
        )
        blueprint.add_url_rule("/", view_func=redirect_fn)
        self.flask.register_blueprint(blueprint)

    def send_file(self, path: str) -> object:
        return send_from_directory(path=path, directory=self._static_rootpath)



# pylint: disable=anomalous-backslash-in-string
"""
Description
-----------

The :class:`flasket.middleware.gunicorn.FlasketCLI` class will enable you to quickly create
and run a Flasket application through the Gunicorn middleware.

It is composed of a :class:`~flasket.middleware.gunicorn.FlasketCmdline` and a
:class:`~flasket.cli.FlasketSettings`. This will enable the following default command line
arguments to be proposed:

.. program:: flasket

.. option:: -l HOST, --listen HOST

   The ip to listen on (default: localhost)

.. option:: -p PORT, --port PORT

   The port to listen on (default: 8080)

.. option:: -c CFGFILE, --cfgfile CFGFILE

   Use CFGFILE as configuration file, otherwise first file found in
   search path is used. (default search path: (...))

.. option:: --ui, --no-ui

   Enable the OpenAPI UI. Disable with :option:`--no-ui`. (default: enabled)

.. option::  -w WORKERS, --workers WORKERS

   Number of thread workers. (default: 0. If 0, cpu to use is (cpu_count * 2) with a maximum
   of 8; if negative, cpu to use is (cpu_count * 2) with no maximum.)

.. option::  --pidfile FILE

   A filename to use for the PID file. (default: none)

A configuration dictionary will be built from the command line arguments, the configuration
file if specified and found, and the Flasket defaults for missing parameters. This
configuration dictionary will be fed to the Flasket application on startup.

Classes
-------

FlasketCmdline
^^^^^^^^^^^^^^

.. autoclass:: FlasketCmdline
  :show-inheritance:
  :members: __call__, add_arguments, transform

FlasketCLI
^^^^^^^^^^

.. autoclass:: FlasketCLI
  :show-inheritance:
  :members: run

Example\s
---------

Default CLI
^^^^^^^^^^^

.. code-block:: python

  from flasket.middleware.gunicorn import FlasketCLI

  if __name__ == "__main__":
      FlasketCLI().run()
"""
import argparse
import logging
import multiprocessing
import typing as t

from attr import define, field

from flask.logging import default_handler
from gunicorn.app.base import BaseApplication
from torxtools.pathtools import expandpath

from .. import cli
from .. import Application


class GunicornApplication(BaseApplication):
    @staticmethod
    def _get_cpu_workers(workers: int) -> int:
        """
        Return the number of workers gunicorn should start with.
        Impose a max limit to avoid overloading the host.

        Parameters
        ----------
        workers: int

            number of workers configured

        Returns
        -------
        int
        """
        if workers > 0:
            return workers

        # Do not overload the host with a default value too large
        # if it's a container, it could report host cpus
        # TODO: respect OMP_NUM_THREADS / OMP_THREAD_LIMIT variables
        cpus = max((multiprocessing.cpu_count() * 2), 1)
        if workers == 0:
            return min(cpus, 8)
        return cpus

    # pylint: disable=abstract-method
    # http://docs.gunicorn.org/en/stable/custom.html
    def __init__(self, *args, cfg: t.Dict, rootpath: str = None, **kwargs) -> None:
        """
        Initialize a Flasket application, and configure Gunicorn

        Parameters
        ----------
        cfg: dict

            dictionary containing configuration for the application

        rootpath: str

            location from where to load packages
        """
        self._flasket = Application(*args, cfg=cfg, rootpath=rootpath, **kwargs)

        # Rework the configuration
        cfg = self._flasket.config["server"]
        cfg["workers"] = self._get_cpu_workers(cfg["workers"])
        self.options = {
            "bind": "%s:%s" % (cfg["listen"], cfg["port"]),
            "pidfile": expandpath(cfg.get("pidfile")),
            "workers": cfg["workers"],
            "accesslog": "/dev/null",
            "errorlog": "/dev/null",
        }

        self._flasket.logger.info(f"Starting {cfg['workers']} workers...")
        super().__init__()

        # Use the Flask logger to log identically between flask and gunicorn
        # We have to specify a null FileStream handler, we'll remove it on first call
        # by using the _gunicorn identifier
        logger = logging.getLogger("gunicorn.error")
        default_handler._gunicorn = False
        logger.addHandler(default_handler)

    def load_config(self) -> None:
        """
        Load Gunicorn configuration ourselves.
        """
        # Pass the configuration down to gunicorn.app.base.BaseApplication
        cfg = {k: v for k, v in self.options.items() if k in self.cfg.settings and v is not None}
        for k, v in cfg.items():
            self.cfg.set(k.lower(), v)

    def load(self) -> None:
        """
        Return our application instance.
        """
        return self._flasket


class FlasketCmdline(cli.FlasketCmdline):
    """
    Class that defines and parses the command line arguments for Gunicorn middleware FlasketCLI.

    Parameters
    ----------
    default_cfg: dict, default: :meth:`flasket.defaults.default_configuration`

        Dictionary containing the defaults for the command line arguments.
        Passed value will be merged with the default factory configuration.
    """

    def add_arguments(self, parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
        """
        Adds Gunicorn middleware specific command line arguments :option:`--workers`, :option:`--pidfile`
        to the ArgumentParser.

        Parameters
        ----------
        parser: argparse.ArgumentParser

            a parser on which `add_argument` will be called

        Returns
        -------
        argparse.ArgumentParser
        """
        defaults = self._default_cfg

        workers = defaults["server"]["workers"]
        pidfile = defaults["server"]["pidfile"]
        if pidfile is None:
            pidfile = "none"

        # fmt: off
        parser.add_argument("-w", "--workers", metavar="WORKERS", type=int,
            help=f"Number of thread workers. (default: {workers}. If 0, cpu to use is (cpu_count * 2) with a maximum of 8; if negative, cpu to use is (cpu_count * 2) with no maximum.)")
        parser.add_argument("--pidfile", metavar="FILE", type=str,
            help=f"A filename to use for the PID file. (default: {pidfile})")
        # fmt: on
        return parser


@define(kw_only=True)
class FlasketCLI(cli.ABCFlasketCli):
    """
    Gunicorn middleware client class.

    Parameters
    ----------
    default_cfg: dict, default: :meth:`flasket.defaults.default_configuration`

        Dictionary containing the defaults for the command line arguments and configuration.
        Passed value will be merged with the default factory configuration.

    settings_cfg: any, default: :class:`flasket.cli.FlasketSettings`

        Dictionary, or Callable that returns a dictionary for the settings arguments.

    cmdline_cfg: any, default: :class:`flasket.middleware.gunicorn.FlasketCmdline`

        Dictionary, or Callable that returns a dictionary for the command line arguments.
    """

    # fmt: off
    _cmdline_cfg: any = field(default=FlasketCmdline)
    _forced_cfg: t.Dict = field(default={"server": {"debug": False, "production": True, "gunicorn": True,}}, init=False)
    # fmt: on

    def _run(self, *args, cfg: t.Dict, rootpath: str = None, **kwargs) -> None:
        """
        Run the middleware

        Parameters
        ----------

        cfg: dict
            dictionary containing configuration for the application

        rootpath: str, default: None

            path of root location: where the 'api', 'htdocs' and 'app' directories will be served
        """
        GunicornApplication(*args, cfg=cfg, rootpath=rootpath, **kwargs).run()


class LocalCmdline(FlasketCmdline):
    def add_arguments(self, parser):
        super().add_arguments(parser)
        # fmt: off
        parser.add_argument("rootpath", default=None, nargs="?",
            help="root path of the server.")
        # fmt: on
        return parser


def main():
    FlasketCLI(cmdline_cfg=LocalCmdline).run()


if __name__ == "__main__":
    main()



import collections
import os
import sys
import traceback
import typing as t

from copy import deepcopy

import yaml

from attrs import define, field

from boltons.fileutils import iter_find_files
from connexion import FlaskApp
from connexion.exceptions import InvalidSpecification
from connexion.resolver import RestyResolver

from .. import utils as dicts
from . import Handler, handler

# TODO: Allow dynamic reload of api


class AutoResolver(RestyResolver):
    def resolve_operation_id_using_rest_semantics(self, operation):
        retval = super().resolve_operation_id_using_rest_semantics(operation)
        retval = retval.replace("api.api.", "api.", 1)
        return retval


def _load_yamlfiles(files: t.List[str]) -> t.Dict:
    """
    For every yaml file, load the file and merge with previous

    Parameters
    ----------
    files: List[str]
        list of files to parse and merge

    Returns
    -------
    dict:
        merged dictionary
    """
    try:
        data = {}
        for f in files:
            with open(f, "r", encoding="UTF-8") as fd:
                newdata = yaml.safe_load(fd)
            if newdata:
                data = dicts.deepmerge(data, newdata)
        return data
    except yaml.scanner.ScannerError as err:
        raise type(err)(
            f"yaml scanner error: {err.problem} in '{err.problem_mark.name}', at line {err.problem_mark.line}"
        ) from err
    except yaml.parser.ParserError as err:
        raise type(err)(
            f"yaml parser error: {err.problem} in '{err.problem_mark.name}', at line {err.problem_mark.line}"
        ) from err
    except Exception as err:
        raise type(err)(f"{str(err)} in file {f}") from err


def _find_yamlfiles(paths: t.List[str], logger: object) -> collections.OrderedDict:
    """
    Find all yaml files in paths, returns a merged dict

    Parameters
    ----------
    paths: List[str]
        list of paths to search

    Returns
    -------
    dict:
        merged dictionary
    """
    # Make paths unique but keep order
    paths = list(dict.fromkeys(paths))

    # Merge all yml files from the directory
    files = []
    for path in paths:
        if os.path.exists(path):
            if logger:
                logger.info(f"Reading OpenAPI files from '{path}'...")
            files.extend(iter_find_files(path, ["*.yml", "*.yaml"]))

    # Load the contents
    data = _load_yamlfiles(files)

    # sort the keys in /components/schemas, since the UI uses that order
    data["components"]["schemas"] = collections.OrderedDict(
        sorted((data["components"]["schemas"]).items(), key=lambda s: s[0].lower())
    )
    return data


def _remove_path_if_key(specs: t.Dict, key: str, value: str) -> t.Dict:
    """
    For every API path in specs, remove it if we find specs[...][key] = value

    Parameters
    ----------
    specs: Dict
        OpenAPI specs

    key: str
        Key to look for

    value: str
        Value to look for

    Returns
    -------
    dict:
        OpenAPI specs without the removed keys
    """
    rv = deepcopy(specs)
    for path in specs.get("paths", {}).keys():
        for method in specs["paths"][path].keys():
            if key not in specs["paths"][path][method]:
                continue
            if specs["paths"][path][method][key] == value:
                del rv["paths"][path][method]
    return rv


@handler
@define(kw_only=True)
class ApiHandler(Handler):
    _connexion = field(init=False, default=None)
    _api_rootpath: str = field(init=False, default=None)

    @property
    def prefix(self) -> str:
        return "/api"

    def __attrs_post_init__(self) -> None:
        self._api_rootpath = os.path.join(self.rootpath, "api")
        if not os.path.exists(self._api_rootpath):
            self.logger.warning("Directory 'api' in root path does not exist.")

        # Build specs from yml files, and remove debug endpoints
        specpath = os.path.abspath(os.path.join(os.path.dirname(__file__), "../api"))
        specs = _find_yamlfiles([specpath, self._api_rootpath], self.logger)
        if not self.debug:
            specs = _remove_path_if_key(specs, "x-debug", True)

        # https://connexion.readthedocs.io/en/latest/
        # TODO: When spec "servers.url" is set, then we should
        # mangle parameters passed here to handle paths prefixed by '/api'
        self._connexion = FlaskApp(
            __name__,
            specification_dir="api",
        )
        try:
            self._connexion.add_api(
                specs,
                # base_path='/api',
                resolver=AutoResolver("api", "get"),
                strict_validation=True,
                validate_responses=True,
                options={
                    "openapi_spec_path": "/api/openapi.json",
                    "swagger_url": "/api/ui",
                    "swagger_ui": self.config["server"]["ui"],
                    "swagger_spec": self.config["server"]["ui"],
                },
            )
        except InvalidSpecification as err:
            raise type(err)(f"OpenAPI error: {err.message}") from err
        except Exception:
            # TODO: improve on this (make import error for example)
            print(traceback.print_exc(), file=sys.stderr)
            raise

        if self.config["server"]["ui"]:
            self.logger.info("Swagger UI is available at: %s%s", self.sitename, "/api/ui")
        else:
            self.logger.info("Swagger UI not started")

    @property
    def flask(self):
        return self._connexion.app

    def __call__(self, environ: dict, start_response):
        return self._connexion(environ, start_response)



from .application import Application
from .clients import client
from .templates import template_global
from .wrappers import endpoint, require_http_same_origin

__all__ = ["Application", "client", "template_global", "endpoint", "require_http_same_origin"]



# pylint: disable=anomalous-backslash-in-string
"""
Description
-----------

The :class:`flasket.middleware.flask.FlasketCLI` class will enable you to quickly create
and run a Flasket application through the Flask middleware.

It is composed of a :class:`~flasket.middleware.flask.FlasketCmdline` and a
:class:`~flasket.cli.FlasketSettings`. This will enable the following default command line
arguments to be proposed:

.. program:: flasket-dev

.. option:: -l HOST, --listen HOST

   The ip to listen on (default: localhost)

.. option:: -p PORT, --port PORT

   The port to listen on (default: 8080)

.. option:: -c CFGFILE, --cfgfile CFGFILE

   Use CFGFILE as configuration file, otherwise first file found in
   search path is used. (default search path: (...))

.. option:: --ui, --no-ui

   Enable the OpenAPI UI. Disable with :option:`--no-ui`. (default: enabled)

.. option:: --debug, --no-debug

  Enable debug mode. Disable with :option:`--no-debug`. (default: disabled)

A configuration dictionary will be built from the command line arguments, the configuration
file if specified and found, and the Flasket defaults for missing parameters. This
configuration dictionary will be fed to the Flasket application on startup.

Classes
-------

FlasketCmdline
^^^^^^^^^^^^^^

.. autoclass:: FlasketCmdline
  :show-inheritance:
  :members: __call__, add_arguments, transform

FlasketCLI
^^^^^^^^^^

.. autoclass:: FlasketCLI
  :show-inheritance:
  :members: run

Example\s
---------

Default CLI
^^^^^^^^^^^

.. code-block:: python

  from flasket.middleware.flask import FlasketCLI

  if __name__ == "__main__":
      FlasketCLI().run()

"""
import argparse
import typing as t

from attr import define, field

from .. import cli
from .. import Application


class FlasketCmdline(cli.FlasketCmdline):
    """
    Class that defines and parses the command line arguments for Flask middleware FlasketCLI.

    Parameters
    ----------

    default_cfg: dict, default: :meth:`flasket.defaults.default_configuration`

        Dictionary containing the defaults for the command line arguments.
        Passed value will be merged with the default factory configuration.
    """

    def add_arguments(self, parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
        """
        Adds Flask middleware specific command line arguments :option:`--debug`, :option:`--no-debug`
        to the ArgumentParser.

        Parameters
        ----------
        parser: argparse.ArgumentParser

            a parser on which `add_argument` will be called

        Returns
        -------
        argparse.ArgumentParser
        """
        defaults = self._default_cfg
        debug = {True: "enabled", False: "disabled"}[defaults["server"]["debug"]]

        # fmt: off
        parser.add_argument("--debug", action="store_true", default=None,
            help=f"Enable debug mode. Disable with --no-debug. (default: {debug})")
        parser.add_argument("--no-debug", action="store_false", default=None, dest="debug",
            help=argparse.SUPPRESS)
        # fmt: on
        return parser


@define(kw_only=True)
class FlasketCLI(cli.ABCFlasketCli):
    """
    Flask middleware client class.

    Parameters
    ----------
    default_cfg: dict, default: :meth:`flasket.defaults.default_configuration`

        Dictionary containing the defaults for the command line arguments and configuration.
        Passed value will be merged with the default factory configuration.

    settings_cfg: any, default: :class:`flasket.cli.FlasketSettings`

        Dictionary, or Callable that returns a dictionary for the settings arguments.
        Defaults to :class:`flasket.cli.FlasketSettings`

    cmdline_cfg: any, default: :class:`flasket.middleware.flask.FlasketCmdline`

        Dictionary, or Callable that returns a dictionary for the command line arguments.
    """

    # fmt: off
    _cmdline_cfg: any = field(default=FlasketCmdline)
    _forced_cfg: t.Dict = field(default={"server": {"production": False, "flask": True,}}, init=False)
    # fmt: on

    def _run(self, *args, cfg: t.Dict, rootpath: str = None, **kwargs) -> None:
        """
        Runs the middleware

        Parameters
        ----------
        cfg: dict

            dictionary containing configuration for the application

        rootpath: str, default: None

            path of root location: where the 'api', 'htdocs' and 'app' directories will be served
        """
        app = Application(*args, cfg=cfg, rootpath=rootpath, **kwargs)
        app.run(host=app.host, port=app.port, debug=app.debug, use_reloader=False, **kwargs)


class LocalCmdline(FlasketCmdline):
    def add_arguments(self, parser):
        super().add_arguments(parser)
        # fmt: off
        parser.add_argument("rootpath", default=None, nargs="?",
            help="root path of the server.")
        # fmt: on
        return parser


def main():
    FlasketCLI(cmdline_cfg=LocalCmdline).run()


if __name__ == "__main__":
    main()
